import json
from typing import Literal
from .postgres_interface import PostgreSQL
from .schema_objects import SQLColumn, SQLSchema



ALTER_COL_OP = Literal["DROP", "ADD", "ALTER", "RENAME COLUMN", "RENAME TABLE"]
POSTGIS_TYPES = Literal["integer", "smallint", "bigint", "real", "double precision", "decimal", "numeric", "smallserial", 
                        "serial", "bigserial", "text", "timestamp with time zone", "timestamp", "date", "time with time zone", 
                        "time", "interval", "uuid", "json", "jsonb", "geometry", "boolean"]

POSTGIS_TYPE_MAP = {"integer": "integer", "smallint": "smallint", "bigint": "bigint", 
                    "real": "real", "double precision": "double precision", "decimal": "decimal", "numeric": "numeric", 
                    "smallserial": "smallserial", "serial": "serial", "bigserial": "bigserial", "geometry": "geometry", "text": "text", 
                    "timestamp": "timestamp", "timestamp with time zone": "timestamp with time zone", "boolean": "boolean",
                    "date": "date", "time with time zone": "time with time zone", "time": "time", "interval": "interval",
                    "uuid": "uuid DEFAULT gen_random_uuid()", "json": "json", "jsonb": "jsonb"}

class PostGIS(PostgreSQL):
    def __init__(self, db_name: str, username: str, password: str, host="localhost", port=5432, type_map=POSTGIS_TYPE_MAP):
        super().__init__(db_name, username, password, host, port, type_map)





from typing import Literal, TypedDict

TYPES = Literal["integer", "smallint", "bigint", "real", "double precision", "decimal", "numeric", "smallserial", 
                "serial", "bigserial", "text", "timestamp with time zone", "timestamp", "date", "time with time zone", 
                "time", "interval", "uuid", "json", "jsonb", "geometry"]

class ForeignKey(TypedDict):
    references_table:str
    references_col:str


class SQLColumn:
    __slots__ = ('name', 'data_type', 'position', 'nullable', 'is_primary_key', 'foreign_key', 'is_unique', 'check_constraint')
    def __init__(self, name:str, data_type:TYPES, position:int = 0, 
                 nullable=True, is_unique:bool=False, is_primary_key:bool = False, 
                 foreign_key:ForeignKey = None, check_constraint:str=""):
        
        self.name:str = name
        self.data_type:TYPES = data_type
        self.position:int = position
        self.nullable:bool = nullable
        self.is_unique:bool = is_unique
        self.is_primary_key:bool = is_primary_key
        self.foreign_key:ForeignKey = foreign_key
        self.check_constraint:str = check_constraint

    def __repr__(self) -> str:
        text = f"{self.name} (data_type: {self.data_type}, position: {self.position}, nullable: {self.nullable}, unique: {self.is_unique}, "
        if self.is_primary_key:
            text += f"Primary Key, "
        if self.foreign_key:
            text += f"Foreign Key: references {self.foreign_key['refences_table']} ({self.foreign_key['references_col']}, "
        if self.check_constraint:
            text += f"Check Constraint: {self.check_constraint}, "
        
        text = text[:-2] + ")"
        return text

class SQLSchema:
    def __init__(self, sql_cols:list[SQLColumn]):
        self.__schema:dict[str, SQLColumn] = {row.name : row for row in sql_cols}

        self.validate_schema()

    @property
    def schema_map(self):
        return self.__schema
    @schema_map.setter
    def schema_map(self, input):
        self.__schema = input
    
    @property
    def col_name_list(self):
        return [col.name for key, col in self.schema_map.items()]
    @property
    def positions_list(self):
        return [col.position for key, col in self.schema_map.items()]
    @property
    def col_positions(self):
        return {key:col.position for key, col in self.schema_map.items()}
    
    def order_by_loc(self):
        for pos, col_name in enumerate(self.schema_map.keys()):
            self.schema_map[col_name].position = pos + 1

    def order_by_pos_id(self):
        temp_schema = {}
        sort_pos = sorted(self.col_positions.items(), key=lambda x: x[1])
        for col in sort_pos:
            temp_schema[col[0]] = self.schema_map[col[0]]
        
        self.schema_map = temp_schema
        
    def add_column(self, col:SQLColumn):
        self.schema_map[col.name] = col

    def drop_column(self, col_name):
        del self.schema_map[col_name]
    
    def filter_columns(self, col_name_list:list):
        extract_map = {}
        for col_name in col_name_list:
            extract_map[col_name] = self.get_col(col_name)
        self.schema_map = extract_map
        self.order_by_loc()
    
    def validate_schema(self):
        pos_list = self.positions_list
        if sorted(pos_list) == list(range(min(pos_list), max(pos_list)+1)):
            self.order_by_pos_id()
        else:
            self.order_by_loc()
    
    def get_col(self, col_name:str) -> SQLColumn:
        return self.schema_map[col_name]
    
    def __repr__(self) -> str:
        text = ""
        for key, val in self.schema_map.items():
            text += f"{val}\n"
        return text





from typing import Protocol, Literal
import psycopg2
from .schema_objects import SQLSchema, SQLColumn

ALTER_COL_OP = Literal["DROP", "ADD", "ALTER", "RENAME COLUMN", "RENAME TABLE"]
DB_TYPE_LIST = Literal["integer", "smallint", "bigint", "real", "double precision", "decimal", "numeric", "text", 
                         "timestamp with time zone", "timestamp", "date", "time with time zone", "time", "interval", 
                         "uuid", "json", "jsonb", "geometry"]

class Database(Protocol):

    @property
    def db_conn(self) -> psycopg2.extensions.connection:
        ...

    def table_from_records(self, table_name:str, table_records:list[dict], col_overrides:list[SQLColumn]=[], schema_override:SQLSchema=None, postgresql_schema:str="public") -> None:
        ...

    def table_append_records(self, table_name:str, table_records:list[dict[str]], postgresql_schema:str="public") -> None:
        ...
    
    def update_with_unique_records(self, table_name:str, records:list[dict], unique_key_col:str, postgresql_schema:str="public") -> None:
        ...

    def delete_rows(self, table_name:str, where_clause:str, postgresql_schema:str="public") -> None:
        ...

    def append_csv_to_table(self, table_name:str, csv_path:str) -> None:
        ...
    
    def records_to_csv(self, csv_name:str, csv_path:str) -> None:
        ...

    def create_blank_table(self, table_name:str, schema:SQLSchema, postgresql_schema:str="public") -> None:
        ...
    
    def csv_to_records(self, csv_path:str) -> list[dict]:
        ...

    def csv_to_table(self, csv_path:str, col_overrides:list[SQLColumn]=[], schema_override:SQLSchema=None, postgresql_schema:str = "public") -> None:
        ...

    def table_to_records(self, table_name:str, columns:str="*", where_clause:str="", postgresql_schema:str="public") -> list[dict]:
        ...
    
    def table_to_csv(self, table_name:str, save_path:str=".", columns:str="*", where_clause:str="", postgresql_schema:str="public") -> None:
        ...

    def drop_table(self, table_name:str) -> None:
        ...
    
    def delete_all_records(self, table_name:str, postgresql_schema:str="public") -> None:
        ...

    def add_column(self, table_name:str, col_name:str, data_type:DB_TYPE_LIST, postgresql_schema:str="public") -> None:
        ...
    
    def drop_column(self, table_name:str, col_name:str, postgresql_schema:str="public") -> None:
        ...
    
    def rename_column(self, table_name:str, col_name:str, new_col_name:str, postgresql_schema:str="public") -> None:
        ...

    def rename_table(self, table_name:str, new_table_name:str, postgresql_schema:str="public") -> None:
        ...

    def list_tables(self, postgresql_schema:str="public") -> list[str]:
        ...

    def get_schema(self, table_name:str, postgresql_schema:str="public") -> SQLSchema:
        ...

    def get_table(self, table_name:str, columns:str="*", where_clause:str="", postgresql_schema:str="public") -> list[tuple]:
        ...

    def evaluate_schema(self, records:list[dict], col_overrides:list[SQLColumn]=[]) -> SQLSchema:
        ...

    def select_table_statement(self, table_name:str, columns:str="*", where_clause:str="") -> str:
        ...

    def create_update_statement(self, table_name:str, record_row:dict, where_clause:str) -> str:
        ...

    def create_delete_statement(self, table_name:str, where_clause:str) -> str:
        ...

    def insert_into_table_statement(self, table_name:str, schema:SQLSchema) -> str:
        ...

    def create_table_statement(self, table_name:str, schema:SQLSchema) -> str:
        ...

    def infer_type(self, val:str) -> str:
        ...









from typing import Literal
import sqlite3
import csv
import pathlib
import psycopg2


#TODO
# Update Table function
# External event listener
# SQL Functions and Triggers
# Postgres TYPEDEF interface
# Expose more DBMS capabilities

POSTGRES_TYPES = {"integer": "integer", "smallint": "smallint", "bigint": "bigint", 
                  "float4": "real", "float8": "double precision", "decimal": "decimal", "numeric": "numeric", 
                  "string": "text", 
                  "geometry": "text", 
                  "epoch_seconds": "integer", "epoch_milliseconds": "real", "timestamp_tz": "timestamp with time zone", "timestamp": "timestamp",
                  "date": "date", "time_tz": "time with time zone", "time": "time", "interval": "interval",
                  "id": "uuid", "json": "json", "jsonb": "jsonb"}

SQLITE_TYPES = {"integer": "INTEGER", "smallint": "INTEGER", "bigint": "INTEGER", 
                "float4": "REAL", "float8": "REAL", "decimal": "REAL", "numeric": "REAL", 
                "string": "TEXT", 
                "geometry": "TEXT", 
                "epoch_seconds": "INTEGER", "epoch_milliseconds": "REAL", "timestamp_tz": "TEXT", "timestamp": "TEXT", 
                "date": "TEXT", "time_tz": "TEXT", "time": "TEXT", "interval": "TEXT",
                "id": "id", "json": "TEXT", "jsonb": "TEXT"}


POSGRES_BINDING = "%s"
SQLITE_BINDING = "?"

ALTER_COL_OP = Literal["DROP", "ADD", "ALTER", "RENAME COLUMN", "RENAME TABLE"]
DB_TYPE_LIST = Literal["integer", "smallint", "bigint", "float4", "float8", "decimal", "numeric", "string", "geometry", "timestamp_tz", "timestamp",
                       "date", "time_tz", "time", "id", "json", "jsonb"]

class DatabaseInterface:
    def __init__(self):
        self.__db_conn = None
        self.__type_map = None
        self.__binding_char = None
    
    @property
    def db_conn(self):
        return self.__db_conn
    
    def set_sqlite_connection(self, db_path:str):
        self.__db_conn = sqlite3.Connection(db_path)
        self.__type_map = SQLITE_TYPES
        self.__binding_char = SQLITE_BINDING
    
    def set_postgres_connection(self, db_name:str, username:str, password:str, host="localhost", port=5432, postgis=False):
        self.__db_conn = psycopg2.connect(database=db_name, user=username, password=password, host=host, port=port)
        self.__type_map = POSTGRES_TYPES
        self.__binding_char = POSGRES_BINDING
        if postgis:
            self.__type_map["geometry"] = "geometry"
        else:
            self.__type_map["geometry"] = "text"
        

    def list_tables(self):
        """Returns a list of tables located within the connected database."""
        if type(self.db_conn) == sqlite3.Connection:
            cur = self.db_conn.cursor()
            res = cur.execute("SELECT name FROM sqlite_master").fetchall()
            return [table[0] for table in res]
        elif type(self.db_conn) == psycopg2.extensions.connection:
            cur = self.db_conn.cursor()
            cur.execute("SELECT table_name FROM information_schema.tables WHERE table_schema = 'public'")
            return [table[0] for table in cur.fetchall()]

    def __ambiguous_type(self, records, col_name):
        for row in records:
            if row[col_name]:
                return self.__infer_type(row[col_name])
        return self.__type_map["string"]

    def __infer_type(self, val:str):
        if type(val) == int or type(val) == float:
            if type(val) == int:
                dtype = self.__type_map["integer"]
            elif type(val) == float:
                dtype = self.__type_map["float8"]
        else:
            val:str
            split = val.split(".")
            if len(split) == 2:
                if split[0].isnumeric() and split[1].isnumeric():
                    if len(split[1]) > 8:
                        dtype = self.__type_map["float8"]
                    else:
                        dtype = self.__type_map["float4"]
                else:
                    dtype = self.__type_map["string"]
            elif val.isnumeric():
                if val == "0":
                    dtype = self.__type_map["numeric"]
                else:
                    dtype = self.__type_map["integer"]
            else:
                dtype = self.__type_map["string"]
            
        return dtype
    
    def __infer_schema(self, records:list[dict], schema_override:list[dict]=[{"col_name": "", "data_type": ""}]):
        schema = []
        top_row:dict = records[0]
        col_overrides = tuple(col['col_name'] for col in schema_override)
        
        position = 1
        for col_name, value in top_row.items():
            if col_name in col_overrides:
                i = col_overrides.index(col_name)
                schema.append({"position": position, "col_name": col_name, "data_type": self.__type_map[schema_override[i]["data_type"]]})
            else:
                if value:
                    data_type = self.__infer_type(value)
                else:
                    data_type = self.__ambiguous_type(records, col_name)
                schema.append({"position": position, "col_name": col_name, "data_type": data_type})
            position += 1
    
        return schema
    
    def __create_table_statement(self, table_name:str, table_schema:list[dict]):
        statement = f"CREATE TABLE IF NOT EXISTS {table_name} ("
        for col in table_schema:
            statement += f"{col['col_name']} {col['data_type']}, "
        statement = statement[:-2] + f")"
        return statement
    
    def __select_table_statement(self, table_name:str, columns:str="*", where_clause:str=""):
        statement = f"SELECT {columns} FROM {table_name}"
        if where_clause:
            statement += f" {where_clause}"
        return statement
    
    def __insert_into_table_statement(self, table_name:str, table_schema:list[dict]):
        statement = f"INSERT INTO {table_name} ("
        bindings = ""
        for col in table_schema:
            statement += f"{col['col_name']}, "
            bindings += f"{self.__binding_char}, "
        statement = statement[:-2] + f") VALUES (" + bindings[:-2] + ")"
        return statement
    
    def __update_table_statement(self):
        pass

    def get_schema(self, table_name:str):
        """Returns a list of dictionaries representing information about a table's schema.\n{"position": 1, "col_name": col1, "data_type": integer}"""
        if type(self.db_conn) == sqlite3.Connection:
            cur = self.db_conn.cursor()
            res = cur.execute(f"PRAGMA table_info({table_name})").fetchall()
            return [{"position": (row[0] + 1), "col_name": row[1], "data_type": row[2]} for row in res]
        elif type(self.db_conn) == psycopg2.extensions.connection:
            cur = self.db_conn.cursor()
            cur.execute(f"SELECT ordinal_position, column_name, data_type FROM information_schema.columns WHERE table_name = '{table_name}'")
            schema = [{"position": row[0], "col_name": row[1], "data_type": row[2]} for row in cur.fetchall()]
            schema = sorted(schema, key=lambda x : x['position'])
            return schema
    
    def get_table(self, table_name:str, columns:str="*", where_clause:str=""):
        """
        Queries the database for a table, returns standard list of tuples representing rows of the table.\n
        table_name 
            - Name of the table\n
        columns (Optional)
            - column name filter separated by commas (col1, col2, col3)
            - Defaults to *\n
        where_clause (Optional)
            - where option for more refine queries
            - must include the "WHERE"
            - Defaults to "" (Empty)
        """
        select_statement = self.__select_table_statement(table_name, columns, where_clause)
        
        if type(self.db_conn) == sqlite3.Connection:
            cur = self.db_conn.cursor()
            res = cur.execute(select_statement).fetchall()
            return res
        elif type(self.db_conn) == psycopg2.extensions.connection:
            cur = self.db_conn.cursor()
            cur.execute(select_statement)
            return cur.fetchall()

    def table_from_records(self, table_name:str, table_records:list[dict], schema_override:list[dict]=[{"col_name": "", "data_type": ""}]):
        """
        Creates a new database table from a list of records.\n
        table_name 
            - Name of the table\n
        table_records
            - A list of dictionaries representing table records
            - [{"col1": "val1", "col2": "val2"}, {"col1": "val1", "col2": "val2"}]\n
        schema_override (Optional)
            - An option to explicitly set a column's datatype
            - Format: [{"col_name": <COLUMN NAME>, "data_type": <DESIRED DATA TYPE>}]
            - DESIRED DATA TYPE: integer, float4, float8, string, timestamp, datetime, geometry
        """
        table_schema = self.__infer_schema(table_records, schema_override)
        create_statement = self.__create_table_statement(table_name, table_schema)
        insert_statement = self.__insert_into_table_statement(table_name, table_schema)
        table_records_sql = [tuple(val for val in row.values()) for row in table_records]

        cur = self.db_conn.cursor()
        cur.execute(create_statement)
        cur.executemany(insert_statement, table_records_sql)
        self.db_conn.commit()
    
    def append_records(self, table_name:str, table_records:list[dict]):
        table_schema = self.get_schema(table_name)
        insert_statement = self.__insert_into_table_statement(table_name, table_schema)
        table_records_sql = [tuple(val for val in row.values()) for row in table_records]

        cur = self.db_conn.cursor()
        cur.executemany(insert_statement, table_records_sql)
        self.db_conn.commit()

    def append_csv(self, table_name:str, csv_path:str):
        csv_path:pathlib.Path = pathlib.Path(csv_path)
        with open(csv_path, newline='') as csv_file:
            reader = csv.DictReader(csv_file)
            records = [row for row in reader]
        self.append_records(table_name, records)

    def csv_to_records(self, csv_path:str):
        csv_path:pathlib.Path = pathlib.Path(csv_path)
        with open(csv_path, newline='') as csv_file:
            reader = csv.DictReader(csv_file)
            records = [row for row in reader]
        
        for row in records:
            for key, val in row.items():
                if val == '':
                    row[key] = None
    
        return records

    def csv_to_table(self, csv_path:str, schema_override:list[dict]=[{"col_name": "", "data_type": ""}]):
        """
        Reads a csv file and migrates into a new database table.\n
        When evaluating the schema, only the top row is checked for data types.
        """
        csv_name = pathlib.Path(csv_path).stem
        records = self.csv_to_records(csv_path)
        if csv_name in self.get_tables():
            print(f"{csv_name} is already a table in the database. You can alternatively use append_csv() if this is intended")
            return
        self.table_from_records(csv_name, records, schema_override)

    def table_to_records(self, table_name:str, columns:str="*", where_clause:str=""):
        """
        Queries the database for a table, returns a list of dictionaries representing table rows.\n
        [{"col1": "val1", "col2": "val2"}, {"col1": "val1", "col2": "val2"}]\n
        table_name 
            - Name of the table\n
        columns (Optional)
            - column name filter separated by commas (col1, col2, col3)
            - Defaults to *\n
        where_clause (Optional)
            - where option for more refine queries
            - must include the "WHERE"
            - Defaults to "" (Empty)
        """
        schema = self.get_schema(table_name)
        table_data = self.get_table(table_name, columns, where_clause)

        if columns != "*":
            col_names = columns.split(", ")
            schema = [col for col in schema if col['col_name'] in col_names]

        records = []
        for row in table_data:
            record_row = {}
            for col in range(len(schema)):
                record_row[schema[col]['col_name']] = row[col]
            records.append(record_row)

        return records
    
    def table_to_csv(self, table_name:str, save_path:str=".", columns:str="*", where_clause:str=""):
        """
        Queries the database for a table, then writes its contents into csv.\n
        table_name 
            - Name of the table\n
        save_path
            - Save location of the csv
            - Do not include a filename, table name is used.\n
        columns (Optional)
            - column name filter separated by commas (col1, col2, col3)
            - Defaults to *\n
        where_clause (Optional)
            - where option for more refine queries
            - must include the "WHERE"
            - Defaults to "" (Empty)
        """
        table_records = self.table_to_records(table_name, columns, where_clause)
        headers:dict = table_records[0]
        headers = headers.keys()

        with open(f"{save_path}\\{table_name}.csv", 'w', newline='') as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=headers)
            writer.writeheader()
            writer.writerows(table_records)
            
    def drop_table(self, table_name):
        cur = self.db_conn.cursor()
        cur.execute(f"DROP TABLE IF EXISTS {table_name}")
        self.db_conn.commit()

    def alter_table(self, table_name, col_name, alter_col_op:ALTER_COL_OP, data_type:DB_TYPE_LIST="", new_col_name:str="", new_table_name:str=""):
        cur = self.db_conn.cursor()
        if alter_col_op == "ADD":
            op = f"ADD {col_name} {self.__type_map[data_type]}"
        elif alter_col_op =="DROP":
            op = f"DROP COLUMN {col_name}"
        elif alter_col_op == "RENAME COLUMN" and new_col_name:
            op = f"RENAME COLUMN {col_name} TO {new_col_name}"
        elif alter_col_op == "RENAME TABLE" and new_table_name:
            op = f"RENAME TO {new_table_name}"
        elif alter_col_op == "ALTER":
            if type(self.db_conn) == sqlite3.Connection:
                raise Exception("Sqlite3 is not capable of modifying column data type")
            op = f"ALTER COLUMN {col_name} {self.__type_map[data_type]}"
        
        sql_statement = f"ALTER TABLE {table_name} {op}"
        cur.execute(sql_statement)
        self.db_conn.commit()





def table_migrate(source_db:DatabaseInterface, dest_db:DatabaseInterface, table_name, dest_table_name=None, schema_override:list[dict]=[{"col_name": "", "data_type": ""}]):
    source_table_records = source_db.table_to_records(table_name)
    if dest_table_name:
        dest_db.table_from_records(dest_table_name, source_table_records, schema_override)
    else:
        dest_db.table_from_records(table_name, source_table_records, schema_override)


from .database import Database
from typing import Literal
import psycopg2
import pathlib
import json
import csv
from datetime import datetime

from .schema_objects import SQLColumn, SQLSchema

TABLE_ACTION = Literal["INSERT", "UPDATE", "DELETE"]
FOR_EACH = Literal["ROW", "STATEMENT"]

POSTGRES_TYPES = Literal["integer", "smallint", "bigint", "real", "double precision", "decimal", "numeric", "text", 
                         "timestamp with time zone", "timestamp", "date", "time with time zone", "time", "interval", 
                         "uuid", "json", "jsonb", "boolean"]
POSTGRES_TYPE_MAP = {"integer": "integer", "smallint": "smallint", "bigint": "bigint", 
                     "real": "real", "double precision": "double precision", "decimal": "decimal", "numeric": "numeric", 
                     "smallserial": "smallserial", "serial": "serial", "bigserial": "bigserial", "geometry": "text", "text": "text", 
                     "timestamp": "timestamp", "timestamp with time zone": "timestamp with time zone", "boolean": "boolean",
                     "date": "date", "time with time zone": "time with time zone", "time": "time", "interval": "interval",
                     "uuid": "uuid DEFAULT gen_random_uuid()", "json": "json", "jsonb": "jsonb"}

class PostgreSQL(Database):
    def __init__(self, db_name:str, username:str, password:str, host="localhost", port=5432, type_map=POSTGRES_TYPE_MAP):
        self.__db_conn = psycopg2.connect(database=db_name, user=username, password=password, host=host, port=port)
        self.__binding_char = "%s"
        self.__type_map = type_map

    @property
    def db_conn(self):
        return self.__db_conn

    def table_from_records(self, table_name:str, table_records:list[dict], col_overrides:list[SQLColumn]=[], schema_override:SQLSchema=None, postgresql_schema:str="public") -> None:
        schema:SQLSchema = schema_override
        if not schema_override:
            schema = self.evaluate_schema(table_records, col_overrides)

        create_statement = self.create_table_statement(f"{postgresql_schema}.{table_name}", schema)
        insert_statement = self.insert_into_table_statement(f"{postgresql_schema}.{table_name}", schema)
        table_records_sql = [tuple(val for val in row.values()) for row in table_records]

        cur = self.db_conn.cursor()
        cur.execute(create_statement)
        cur.executemany(insert_statement, table_records_sql)
        self.db_conn.commit()

    def table_append_records(self, table_name:str, table_records:list[dict[str]], postgresql_schema:str="public") -> None:
        schema = self.get_schema(table_name, postgresql_schema)
        schema.filter_columns(list(table_records[0].keys()))
        
        insert_statement = self.insert_into_table_statement(f"{postgresql_schema}.{table_name}", schema)
        table_records_sql = [tuple(val for val in row.values()) for row in table_records]

        cur = self.db_conn.cursor()
        cur.executemany(insert_statement, table_records_sql)
        self.db_conn.commit()

    def update_with_unique_records(self, table_name:str, records:list[dict], unique_key_col:str, postgresql_schema:str="public") -> None:
        cur = self.db_conn.cursor()
        for row in records:
            update_statement = self.create_update_statement(f"{postgresql_schema}.{table_name}", row, f"WHERE {unique_key_col}='{row[unique_key_col]}'")
            cur.execute(update_statement)
        self.db_conn.commit()

    def delete_rows(self, table_name:str, where_clause:str, postgresql_schema:str="public") -> None:
        cur = self.db_conn.cursor()
        delete_statement = self.create_delete_statement(f"{postgresql_schema}.{table_name}", where_clause)
        cur.execute(delete_statement)
        self.db_conn.commit()

    def append_csv_to_table(self, table_name:str, csv_path:str) -> None:
        csv_path:pathlib.Path = pathlib.Path(csv_path)
        with open(csv_path, newline='') as csv_file:
            reader = csv.DictReader(csv_file)
            records = [row for row in reader]
        self.table_append_records(table_name, records)

    def records_to_csv(self, records:list[dict], csv_name:str, csv_path:str) -> None:
        headers = records[0].keys()
        with open(f"{csv_path}\\{csv_name}.csv", "w", newline='') as f:
            writer = csv.DictWriter(f, headers)
            writer.writeheader()
            writer.writerows(records)

    def create_blank_table(self, table_name:str, schema:SQLSchema, postgresql_schema:str="public") -> None:
        sql = self.create_table_statement(f"{postgresql_schema}.{table_name}", schema)
        cur = self.db_conn.cursor()
        cur.execute(sql)
        self.db_conn.commit()
        
    def csv_to_records(self, csv_path:str) -> list[dict]:
        csv_path:pathlib.Path = pathlib.Path(csv_path)
        with open(csv_path, newline='') as csv_file:
            reader = csv.DictReader(csv_file)
            records = [row for row in reader]
        
        for row in records:
            for key, val in row.items():
                if val == '':
                    row[key] = None
    
        return records
    
    def csv_to_table(self, csv_path:str, col_overrides:list[SQLColumn]=[], schema_override:SQLSchema=None, postgresql_schema:str = "public") -> None:
        csv_name = pathlib.Path(csv_path).stem
        records = self.csv_to_records(csv_path)
        if csv_name in self.list_tables():
            print(f"{csv_name} is already a table in the database. You can alternatively use append_csv() if this is intended")
            return
        self.table_from_records(csv_name, records, col_overrides, schema_override, postgresql_schema)

    def table_to_records(self, table_name:str, columns:str="*", where_clause:str="", postgresql_schema:str="public") -> list[dict]:
        schema:SQLSchema = self.get_schema(table_name, postgresql_schema)
        table_data = self.get_table(table_name, columns, where_clause, postgresql_schema)

        if columns != "*":
            schema.filter_columns(columns.split(", "))
            
        records = []
        for row in table_data:
            record_row = {}
            for col_name, sql_col in schema.schema_map.items():
                record_row[col_name] = row[sql_col.position - 1]
            records.append(record_row)

        return records

    def table_to_csv(self, table_name:str, save_path:str=".", columns:str="*", where_clause:str="", postgresql_schema:str="public") -> None:
        table_records = self.table_to_records(table_name, columns, where_clause, postgresql_schema)
        headers:dict = table_records[0]
        headers = headers.keys()

        with open(f"{save_path}\\{table_name}.csv", 'w', newline='') as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=headers)
            writer.writeheader()
            writer.writerows(table_records)
    
    def drop_table(self, table_name:str, postgresql_schema:str) -> None:
        cur = self.db_conn.cursor()
        cur.execute(f"DROP TABLE IF EXISTS {postgresql_schema}.{table_name}")
        self.db_conn.commit()
    
    def delete_all_records(self, table_name:str, postgresql_schema:str="public") -> None:
        cur = self.db_conn.cursor()
        cur.execute(f"DELETE FROM {postgresql_schema}.{table_name}")
        self.db_conn.commit()

    def add_column(self, table_name:str, col_name:str, data_type:POSTGRES_TYPES, postgresql_schema:str="public") -> None:
        cur = self.db_conn.cursor()
        sql_statement = f"ALTER TABLE {postgresql_schema}.{table_name} ADD {col_name} {self.__type_map[data_type]}"
        cur.execute(sql_statement)
        self.db_conn.commit()
    
    def drop_column(self, table_name:str, col_name:str, postgresql_schema:str="public") -> None:
        cur = self.db_conn.cursor()
        sql_statement = f"ALTER TABLE {postgresql_schema}.{table_name} DROP COLUMN {col_name}"
        cur.execute(sql_statement)
        self.db_conn.commit()
    
    def rename_column(self, table_name:str, col_name:str, new_col_name:str, postgresql_schema:str="public") -> None:
        cur = self.db_conn.cursor()
        sql_statement = f"ALTER TABLE {postgresql_schema}.{table_name} RENAME COLUMN {col_name} TO {new_col_name}"
        cur.execute(sql_statement)
        self.db_conn.commit()

    def rename_table(self, table_name:str, new_table_name:str, postgresql_schema:str="public") -> None:
        cur = self.db_conn.cursor()
        sql_statement = f"ALTER TABLE {postgresql_schema}.{table_name} RENAME TO {new_table_name}"
        cur.execute(sql_statement)
        self.db_conn.commit()

    def alter_column(self, table_name:str, col_name:str, data_type:POSTGRES_TYPES, postgresql_schema:str="public") -> None:
        cur = self.db_conn.cursor()
        sql_statement = f"ALTER TABLE {postgresql_schema}.{table_name} ALTER COLUMN {col_name} {self.__type_map[data_type]}"
        cur.execute(sql_statement)
        self.db_conn.commit()

    def create_trigger_function(self, function_name:str, function_sql:str) -> None:
        create_function = f"CREATE OR REPLACE FUNCTION {function_name}() RETURNS trigger AS $$"
        create_function += f"BEGIN {function_sql}"
        create_function += "END; $$ LANGUAGE plpgsql;"

        cur = self.db_conn.cursor()
        cur.execute(create_function)
        self.db_conn.commit()

    def create_trigger(self, trigger_name:str, after_action:TABLE_ACTION, for_each:FOR_EACH, table_name:str, call_function:str) -> None:
        trigger_sql = f"CREATE TRIGGER {trigger_name}"
        trigger_sql += f"AFTER {after_action} ON {table_name}"
        trigger_sql += f"FOR EACH {for_each} EXECUTE FUNCTION {call_function}();"

        cur = self.db_conn.cursor()
        cur.execute(trigger_sql)
        self.db_conn.commit()

    def list_tables(self, postgresql_schema:str="public") -> list[str]:
        cur = self.db_conn.cursor()
        sql = f"SELECT table_name FROM information_schema.tables WHERE table_schema = '{postgresql_schema}'"
        cur.execute(sql)
        return [table[0] for table in cur.fetchall()]

    def get_schema(self, table_name:str, postgresql_schema:str="public") -> SQLSchema:
        GET_COL_SCHEMA = f"SELECT column_name, data_type, ordinal_position, is_nullable FROM information_schema.columns WHERE table_schema='{postgresql_schema}' AND table_name='{table_name}'"
        cur = self.db_conn.cursor()
        cur.execute(GET_COL_SCHEMA)
        col_list = []
        for col in cur.fetchall():
            if col[3] == 'YES':
                nullable = True
            else:
                nullable = False
            col_list.append(SQLColumn(col[0], col[1], col[2], nullable))

        schema = SQLSchema(col_list)
        return schema
    
    def get_table(self, table_name:str, columns:str="*", where_clause:str="", postgresql_schema:str="public") -> list[tuple]:
        select_statement = self.select_table_statement(f"{postgresql_schema}.{table_name}", columns, where_clause)
        cur = self.db_conn.cursor()
        cur.execute(select_statement)
        return cur.fetchall()

    def evaluate_schema(self, records:list[dict], col_overrides:list[SQLColumn]=[]) -> SQLSchema:

        def find_non_null_record(records:list[dict], col_name:str):
            for row in records:
                if row[col_name]:
                    return row[col_name]
            return "empty_col_default_to_string"

        schema = []
        top_row:dict = records[0]
        overide_col_list = tuple(col.name for col in col_overrides)

        pos = 1
        for col_name, col_val in top_row.items():
            if col_name in overide_col_list:
                col:SQLColumn = col_overrides[overide_col_list.index(col_name)]
                schema.append(SQLColumn(name=col.name, data_type=self.__type_map[col.data_type], position=pos, 
                                        is_primary_key=col.is_primary_key, foreign_key=col.foreign_key, 
                                        is_unique=col.is_unique, check_constraint=col.check_constraint,
                                        not_null=col.nullable))
            else:
                if col_val:
                    d_type = self.infer_type(col_val)
                else:
                    d_type = self.infer_type(find_non_null_record(records, col_name))
                
                schema.append(SQLColumn(name=col_name, data_type=d_type, position=pos))
            pos += 1
        return SQLSchema(schema)

    def select_table_statement(self, table_name:str, columns:str="*", where_clause:str="") -> str:
        statement = f"SELECT {columns} FROM {table_name}"
        if where_clause:
            statement += f" {where_clause}"
        return statement
    
    def create_update_statement(self, table_name:str, record_row:dict, where_clause:str) -> str:
        statement = f"UPDATE {table_name} SET "
        for col_name, value in record_row.items():
            statement += f"{col_name}='{value}', "
        statement = statement[:-2] + " " + where_clause
        return statement
    
    def create_delete_statement(self, table_name:str, where_clause:str="") -> str:
        statement = f"DELETE FROM {table_name}"
        if where_clause:
            statement += f" {where_clause}"
        return statement
    
    def insert_into_table_statement(self, table_name:str, schema:SQLSchema) -> str:
        statement = f"INSERT INTO {table_name} ("
        bindings = ""
        for col_name, sql_col in schema.schema_map.items():
            statement += f"{col_name}, "
            bindings += f"{self.__binding_char}, "
        statement = statement[:-2] + f") VALUES (" + bindings[:-2] + ")"
        return statement

    def create_table_statement(self, table_name:str, schema:SQLSchema) -> str:
        statement = f"CREATE TABLE IF NOT EXISTS {table_name} ("
        for_keys = ""

        for col_name, sql_col in schema.schema_map.items():
            statement += f"{col_name} {self.__type_map[sql_col.data_type]}"
            if sql_col.is_primary_key:
                statement = statement + f" PRIMARY KEY"
            if sql_col.is_unique and not sql_col.is_primary_key:
                statement = statement + f" UNIQUE"
            if not sql_col.nullable and not sql_col.is_primary_key:
                statement = statement + f" NOT NULL"
            if sql_col.check_constraint:
                statement = statement + f" {sql_col.check_constraint}"
            if sql_col.foreign_key:
                statement = statement + f" REFERENCES {sql_col.foreign_key['refences_table']} ({sql_col.foreign_key['references_col']})"
            statement = statement + ", "
        if for_keys:
            statement += for_keys
        statement = statement[:-2] + f")"
        return statement

    def infer_type(self, val:str) -> str:

        def string_jsonable(x):
            try:
                json.loads(x)
                return True
            except (TypeError, OverflowError, json.decoder.JSONDecodeError):
                return False
        def dict_jsonable(x):
            try:
                json.dumps(x)
                return True
            except (TypeError, OverflowError, json.decoder.JSONDecodeError):
                return False

        if string_jsonable(val):
            return "json"
        elif type(val) == dict and dict_jsonable(val):
            return "json"

        if type(val) == int or type(val) == float:
            if type(val) == int:
                return "integer"
            elif type(val) == float:
                return "decimal"
        elif type(val) == bool:
            return "text"
        elif type(val) == datetime:
            return "timestamp"
        else:
            val:str
            split = val.split(".")
            if len(split) == 2:
                if split[0].isnumeric() and split[1].isnumeric():
                    if len(split[1]) > 8:
                        return "double precision"
                    else:
                        return "real"
                else:
                    return "text"
            elif val.isnumeric():
                if val == "0":
                    return "numeric"
                else:
                    return "integer"
            else:
                return "text"




from .database import Database
import sqlite3
import csv
from datetime import datetime
from typing import Literal
import pathlib

from .schema_objects import SQLColumn, SQLSchema

SQLITE_TYPES = Literal["text", "real", "integer", "numeric"]

SQLITE_TYPE_MAP = {"integer": "integer", "smallint": "integer", "bigint": "integer", 
                   "real": "real", "double precision": "real", "decimal": "real", "numeric": "numeric",  
                   "smallserial": "integer", "serial": "integer", "bigserial": "integer", "geometry": "text", 
                   "text": "text", "timestamp": "text", "timestamp with time zone": "text",
                   "date": "text", "time with timezone": "text", "time": "text", "boolean": "text",
                   "uuid": "text", "interval": "text",  "json": "text", "jsonb": "text"}

class SQLite(Database):
    def __init__(self, sqlite_path:str, foreign_keys_on=False, type_map=SQLITE_TYPE_MAP):
        self.__db_conn = sqlite3.Connection(sqlite_path)
        self.__binding_char = "?"
        self.__type_map = type_map
        if foreign_keys_on:
            cur = self.db_conn.cursor()
            cur.execute("PRAGMA foreign_keys = ON")
            resp = cur.execute("PRAGMA foreign_keys")
            print(f"Foreign Keys for {sqlite_path}:", bool(resp.fetchall()[0][0]))
    
    @property
    def db_conn(self):
        return self.__db_conn
    
    def table_from_records(self, table_name:str, table_records:list[dict], col_overrides:list[SQLColumn]=[], schema_override:SQLSchema=None) -> None:
        schema:SQLSchema = schema_override
        if not schema_override:
            schema = self.evaluate_schema(table_records, col_overrides)

        create_statement = self.create_table_statement(table_name, schema)
        insert_statement = self.insert_into_table_statement(table_name, schema)
        table_records_sql = [tuple(val for val in row.values()) for row in table_records]

        cur = self.db_conn.cursor()
        cur.execute(create_statement)
        cur.executemany(insert_statement, table_records_sql)
        self.db_conn.commit()

    def table_append_records(self, table_name:str, table_records:list[dict]) -> None:
        schema = self.get_schema(table_name)
        schema.filter_columns(list(table_records[0].keys()))
        
        insert_statement = self.insert_into_table_statement(table_name, schema)
        table_records_sql = [tuple(val for val in row.values()) for row in table_records]

        cur = self.db_conn.cursor()
        cur.executemany(insert_statement, table_records_sql)
        self.db_conn.commit()

    def update_with_unique_records(self, table_name:str, records:list[dict], unique_key_col:str) -> None:
        cur = self.db_conn.cursor()
        for row in records:
            update_statement = self.create_update_statement(table_name, row, f"WHERE {unique_key_col}='{row[unique_key_col]}'")
            cur.execute(update_statement)
        self.db_conn.commit()

    def delete_rows(self, table_name:str, where_clause:str) -> None:
        cur = self.db_conn.cursor()
        delete_statement = self.create_delete_statement(table_name, where_clause)
        cur.execute(delete_statement)
        self.db_conn.commit()

    def append_csv_to_table(self, table_name:str, csv_path:str) -> None:
        csv_path:pathlib.Path = pathlib.Path(csv_path)
        with open(csv_path, newline='') as csv_file:
            reader = csv.DictReader(csv_file)
            records = [row for row in reader]
        self.table_append_records(table_name, records)

    def records_to_csv(self, records:list[dict], csv_name:str, csv_path:str) -> None:
        headers = records[0].keys()
        with open(f"{csv_path}\\{csv_name}.csv", "w", newline='') as f:
            writer = csv.DictWriter(f, headers)
            writer.writeheader()
            writer.writerows(records)

    def create_blank_table(self, table_name:str, schema:SQLSchema) -> None:
        sql = self.create_table_statement(table_name, schema)
        cur = self.db_conn.cursor()
        cur.execute(sql)
        
    def csv_to_records(self, csv_path:str) -> list[dict]:
        csv_path:pathlib.Path = pathlib.Path(csv_path)
        with open(csv_path, newline='') as csv_file:
            reader = csv.DictReader(csv_file)
            records = [row for row in reader]
        
        for row in records:
            for key, val in row.items():
                if val == '':
                    row[key] = None
    
        return records
    
    def csv_to_table(self, csv_path:str, col_overrides:list[SQLColumn]=[], schema_override:SQLSchema=None) -> None:
        csv_name = pathlib.Path(csv_path).stem
        records = self.csv_to_records(csv_path)
        if csv_name in self.list_tables():
            print(f"{csv_name} is already a table in the database. You can alternatively use append_csv() if this is intended")
            return
        self.table_from_records(csv_name, records, col_overrides, schema_override)

    def table_to_records(self, table_name:str, columns:str="*", where_clause:str="") -> list[dict]:
        schema:SQLSchema = self.get_schema(table_name)
        table_data = self.get_table(table_name, columns, where_clause)

        if columns != "*":
            schema.filter_columns(columns.split(", "))
            
        records = []
        for row in table_data:
            record_row = {}
            for col_name, sql_col in schema.schema_map.items():
                record_row[col_name] = row[sql_col.position - 1]
            records.append(record_row)

        return records

    def table_to_csv(self, table_name:str, save_path:str=".", columns:str="*", where_clause:str="") -> None:
        table_records = self.table_to_records(table_name, columns, where_clause)
        headers:dict = table_records[0]
        headers = headers.keys()

        with open(f"{save_path}\\{table_name}.csv", 'w', newline='') as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=headers)
            writer.writeheader()
            writer.writerows(table_records)
            
    def drop_table(self, table_name) -> None:
        cur = self.db_conn.cursor()
        cur.execute(f"DROP TABLE IF EXISTS {table_name}")
        self.db_conn.commit()

    def delete_all_records(self, table_name:str) -> None:
        cur = self.db_conn.cursor()
        cur.execute(f"DELETE FROM {table_name}")
        self.db_conn.commit()

    def add_column(self, table_name:str, col_name:str, data_type:SQLITE_TYPES):
        cur = self.db_conn.cursor()
        sql_statement = f"ALTER TABLE {table_name} ADD {col_name} {self.__type_map[data_type]}"
        cur.execute(sql_statement)
        self.db_conn.commit()
    
    def drop_column(self, table_name:str, col_name:str):
        cur = self.db_conn.cursor()
        sql_statement = f"ALTER TABLE {table_name} DROP COLUMN {col_name}"
        cur.execute(sql_statement)
        self.db_conn.commit()
    
    def rename_column(self, table_name:str, col_name:str, new_col_name:str):
        cur = self.db_conn.cursor()
        sql_statement = f"ALTER TABLE {table_name} RENAME COLUMN {col_name} TO {new_col_name}"
        cur.execute(sql_statement)
        self.db_conn.commit()

    def rename_table(self, table_name:str, new_table_name:str):
        cur = self.db_conn.cursor()
        sql_statement = f"ALTER TABLE {table_name} RENAME TO {new_table_name}"
        cur.execute(sql_statement)
        self.db_conn.commit()

    def list_tables(self) -> list[str]:
        cur = self.db_conn.cursor()
        res = cur.execute("SELECT name FROM sqlite_master").fetchall()
        return [table[0] for table in res]

    def get_schema(self, table_name:str) -> SQLSchema:
        cur = self.db_conn.cursor()
        res = cur.execute(f"PRAGMA table_info({table_name})").fetchall()
        if not res:
            print(f"No data returned from table '{table_name}'")
            return
        return SQLSchema([SQLColumn(name=row[1], data_type=row[2], position=(row[0] + 1), nullable=bool(row[3]), is_primary_key=bool(row[5])) for row in res])
    
    def get_table(self, table_name:str, columns:str="*", where_clause:str="") -> list[tuple]:
        select_statement = self.select_table_statement(table_name, columns, where_clause)
        cur = self.db_conn.cursor()
        res = cur.execute(select_statement).fetchall()
        return res
    
    def infer_type(self, val:str) -> str:
        if type(val) == int or type(val) == float:
            if type(val) == int:
                return "integer"
            elif type(val) == float:
                return "real"
        elif type(val) == bool:
            return "text"
        elif type(val) == datetime:
            return "text"
        else:
            val:str
            split = val.split(".")
            if len(split) == 2:
                if split[0].isnumeric() and split[1].isnumeric():
                    return "real"
                else:
                    return "text"
            elif val.isnumeric():
                if val == "0":
                    return "numeric"
                else:
                    return "integer"
            else:
                return "text"
            
    def evaluate_schema(self, records:list[dict], col_overrides:list[SQLColumn]=[]) -> SQLSchema:

        def find_non_null_record(records:list[dict], col_name:str):
            for row in records:
                if row[col_name]:
                    return row[col_name]
            return "empty_col_default_to_string"

        schema = []
        top_row:dict = records[0]
        overide_col_list = tuple(col.name for col in col_overrides)

        pos = 1
        for col_name, col_val in top_row.items():
            if col_name in overide_col_list:
                col:SQLColumn = col_overrides[overide_col_list.index(col_name)]
                schema.append(SQLColumn(name=col.name, data_type=self.__type_map[col.data_type], position=pos, 
                                        not_null=col.nullable, is_primary_key=col.is_primary_key, foreign_key=col.foreign_key))
            else:
                if col_val:
                    d_type = self.infer_type(col_val)
                else:
                    d_type = self.infer_type(find_non_null_record(records, col_name))
                
                schema.append(SQLColumn(name=col_name, data_type=d_type, position=pos))
            pos += 1
        return SQLSchema(schema)
    
    def select_table_statement(self, table_name:str, columns:str="*", where_clause:str="") -> str:
        statement = f"SELECT {columns} FROM {table_name}"
        if where_clause:
            statement += f" {where_clause}"
        return statement
    
    def create_update_statement(self, table_name:str, record_row:dict, where_clause:str) -> str:
        statement = f"UPDATE {table_name} SET "
        for col_name, value in record_row.items():
            statement += f"{col_name}='{value}', "
        statement = statement[:-2] + " " + where_clause
        return statement

    def create_delete_statement(self, table_name:str, where_clause:str="") -> str:
        statement = f"DELETE FROM {table_name}"
        if where_clause:
            statement += f" {where_clause}"
        return statement

    def insert_into_table_statement(self, table_name:str, schema:SQLSchema) -> str:
        statement = f"INSERT INTO {table_name} ("
        bindings = ""
        for col_name, sql_col in schema.schema_map.items():
            statement += f"{col_name}, "
            bindings += f"{self.__binding_char}, "
        statement = statement[:-2] + f") VALUES (" + bindings[:-2] + ")"
        return statement

    def create_table_statement(self, table_name:str, schema:SQLSchema) -> str:
        statement = f"CREATE TABLE IF NOT EXISTS {table_name} ("
        for_keys = ""
        for col_name, sql_col in schema.schema_map.items():
            statement += f"{col_name} {self.__type_map[sql_col.data_type]}"
            if sql_col.is_primary_key:
                statement = statement + f" PRIMARY KEY"
            if not sql_col.nullable:
                statement = statement + f" NOT NULL"
            if sql_col.foreign_key:
                for_keys += f"FOREIGN KEY({col_name}) REFERENCES {sql_col.foreign_key['refences_table']} ({sql_col.foreign_key['references_col']}), "
            statement = statement + ", "
        if for_keys:
            statement += for_keys
        statement = statement[:-2] + f")"
        return statement
    


from database_interface.database import Database
from database_interface.schema_objects import SQLColumn, SQLSchema, ForeignKey
from database_interface.sqlite_interface import SQLite
from database_interface.postgres_interface import PostgreSQL
from database_interface.postgis_interface import PostGIS
